# Copyright 2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ros2cli.node.strategy import add_arguments
from ros2cli.node.strategy import NodeStrategy

from ros2component.api import add_component_arguments
from ros2component.api import container_node_name_completer
from ros2component.api import find_container_node_names
from ros2component.api import load_component_into_container
from ros2component.verb import VerbExtension

from ros2node.api import get_node_names


class LoadVerb(VerbExtension):
    """Load a component into a container node."""

    def add_arguments(self, parser, cli_name):
        add_arguments(parser)
        argument = parser.add_argument(
            'container_node_name', help='Container node name to unload component from'
        )
        argument.completer = container_node_name_completer
        add_component_arguments(parser)
        parser.add_argument(
            '-q', '--quiet', action='store_true', default=False,
            help='Only print component unique IDs and names'
        )

    def main(self, *, args):
        with NodeStrategy(args) as node:
            container_node_names = find_container_node_names(
                node=node, node_names=get_node_names(node=node)
            )
            if args.container_node_name not in [n.full_name for n in container_node_names]:
                return "Unable to find container node '" + args.container_node_name + "'"
            component_uid, component_name = load_component_into_container(
                node=node, remote_container_node_name=args.container_node_name,
                package_name=args.package_name, plugin_name=args.plugin_name,
                node_name=args.node_name, node_namespace=args.node_namespace,
                log_level=args.log_level, remap_rules=args.remap_rules,
                parameters=args.parameters, extra_arguments=args.extra_arguments
            )
            if not args.quiet:
                print("Loaded component {} into '{}' container node as '{}'".format(
                    component_uid, args.container_node_name, component_name
                ))
            else:
                print('{}  {}'.format(component_uid, component_name))



# Copyright 2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import platform
import signal
import uuid

from ros2cli.node.direct import DirectNode
from ros2cli.node.strategy import add_arguments

from ros2component.api import add_component_arguments
from ros2component.api import load_component_into_container
from ros2component.api import run_standalone_container
from ros2component.verb import VerbExtension


class StandaloneVerb(VerbExtension):
    """Run a component into its own standalone container node."""

    def add_arguments(self, parser, cli_name):
        add_arguments(parser)
        add_component_arguments(parser)
        parser.add_argument(
            '-c', '--container-node-name',
            default='standalone_container_' + uuid.uuid4().hex[:12],
            help='Name of the standalone container node to be run'
        )

    def main(self, *, args):
        container = run_standalone_container(container_node_name=args.container_node_name)

        with DirectNode(args) as node:
            try:
                load_component_into_container(
                    node=node, remote_container_node_name=args.container_node_name,
                    package_name=args.package_name, plugin_name=args.plugin_name,
                    node_name=args.node_name, node_namespace=args.node_namespace,
                    log_level=args.log_level, remap_rules=args.remap_rules,
                    parameters=args.parameters, extra_arguments=args.extra_arguments
                )
            except RuntimeError as ex:
                # In case the component fails to load, kill the container.
                print('{}, interrupting container node'.format(ex))
                if platform.system() == 'Windows':
                    container.send_signal(signal.CTRL_C_EVENT)
                else:
                    container.send_signal(signal.SIGINT)

        while container.returncode is None:
            try:
                container.communicate()
            except KeyboardInterrupt:
                # the subprocess will also receive the signal and should shut down
                # therefore we continue here until the process has finished
                pass
        return container.returncode



# Copyright 2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ros2cli.node.strategy import add_arguments
from ros2component.api import get_package_component_types
from ros2component.api import get_registered_component_types
from ros2component.verb import VerbExtension

from ros2pkg.api import get_package_names
from ros2pkg.api import package_name_completer


class TypesVerb(VerbExtension):
    """Output a list of components registered in the ament index."""

    def add_arguments(self, parser, cli_name):
        add_arguments(parser)
        argument = parser.add_argument(
            'package_name', nargs='?', default=None,
            help='Package name to look for registered components in'
        )
        argument.completer = package_name_completer

    def main(self, *, args):
        if args.package_name is not None:
            if args.package_name not in get_package_names():
                return "Unable to find package '" + args.package_name + "'"
            print(*get_package_component_types(package_name=args.package_name), sep='\n')
        else:
            for package_name, component_types in get_registered_component_types():
                print(package_name)
                print(*['  ' + type_name for type_name in component_types], sep='\n')



# Copyright 2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ros2cli.node.strategy import add_arguments
from ros2cli.node.strategy import NodeStrategy

from ros2component.api import container_node_name_completer
from ros2component.api import find_container_node_names
from ros2component.api import get_components_in_container
from ros2component.api import get_components_in_containers
from ros2component.verb import VerbExtension

from ros2node.api import get_node_names


class ListVerb(VerbExtension):
    """Output a list of running containers and components."""

    def add_arguments(self, parser, cli_name):
        add_arguments(parser)
        argument = parser.add_argument(
            'container_node_name', nargs='?', default=None,
            help='Name of the container node to list components from')
        argument.completer = container_node_name_completer
        parser.add_argument(
            '--containers-only', action='store_true',
            help='List found containers nodes only')

    def main(self, *, args):
        with NodeStrategy(args) as node:
            container_node_names = find_container_node_names(
                node=node, node_names=get_node_names(node=node)
            )
            if args.container_node_name is not None:
                if args.container_node_name not in [n.full_name for n in container_node_names]:
                    return "Unable to find container node '" + args.container_node_name + "'"
                if not args.containers_only:
                    ok, outcome = get_components_in_container(
                        node=node, remote_container_node_name=args.container_node_name
                    )
                    if not ok:
                        return f'{outcome} when listing components in {args.container_node_name}'
                    if any(outcome):
                        print(*[
                            f'{component.uid}  {component.name}' for component in outcome
                        ], sep='\n')
            else:
                results = get_components_in_containers(node=node, remote_containers_node_names=[
                    n.full_name for n in container_node_names
                ])
                for container_node_name, (ok, outcome) in results.items():
                    print(container_node_name)
                    if not args.containers_only:
                        if not ok:
                            print(f'{outcome} when listing components')
                            continue
                        if any(outcome):
                            print(*[
                                f'  {component.uid}  {component.name}' for component in outcome
                            ], sep='\n')
